(*
    Project: Component Pascal command-line compiler
    Organization: Oberon Revival team
    License: BSD style license
    References: http://oberonrevival.sourceforge.net/
*)

{$IFDEF FPC}
    {$mode delphi}
    {$smartlink on}
{$ENDIF}

Unit cdev;

Interface

{$ifdef win32}
    {$I consts-win32.inc}
{$endif}
{$ifdef linux}
    {$I consts-linux.inc}
{$endif}

const
    resSuccess = 0;
    resFailed = 1;
    resSymNotFound = 2;

    isdll = 0;
    isintf = 1;
    iskernel = 2;
    ismain = 3;
    
    RES_INVALID_PARAMETER = 1;	// invalid parameter (name or locator)
    RES_NOT_FOUND = 2;	// location or file not found
    RES_ALREADY_EXISTS = 3;	// file already exists
    RES_WRITE_PROTECTION = 4;	// write-protection
    RES_IO_ERROR = 5;	// io error
    RES_ACCESS_DENIED = 6;	// access denied
    
type
    PCompError = ^TCompError;
    TCompError = record
        No,
        col,
        row: LongInt;
    end;

    PCompData = ^TCompData;
    TCompData = record
        pc, dsize: LongInt
    end;

    PErrType = ^ErrType;
    ErrType = record
        error: LongBool;
        mes: array [0..127] of Char;
        location: LongInt
    end;

    TName = array [0..39] of char;
    PName = ^TName;

    (* Compiler symbols *)
    TSetCompilerOptionsProc = procedure (opt: PChar); stdcall;
    TCompileFileProc = function (argpath: PChar; var arg3: LongInt): boolean; stdcall;
    TVersionDecodeProc = procedure (var major, minor, micro: LongInt); stdcall;
    (* BuilderKit symbols *)
    TGetCompilerErrorsProc = function (): LongInt; stdcall;
    TGetCompilerErrorInfoProc = procedure (var error: TCompError); stdcall;
    TGetCompilerStatusProc = procedure (var compdata: TCompData); stdcall;
    TSetDllNameProc = procedure (dllName: PChar); stdcall;
    TAddModuleProc = procedure (name: PChar; opts: LongInt; var root: Pointer); stdcall;
    TModListInitProc = procedure (var root: Pointer); stdcall;
    TSetModuleOptionProc = procedure (var iopts: LongInt; option: LongInt); stdcall;
    TGetLinkerErrorProc = procedure (var error: ErrType); stdcall;
    TGetLinkerStatusProc = procedure (var codesize, filelen: LongInt); stdcall;
    (* Linker symbols *)
    TLinkDllProc = function (m, r: Pointer): boolean; stdcall;
    TLinkExeProc = function (m, r: Pointer; c: boolean): boolean; stdcall;

    function LoadEngine: integer;
    procedure UnloadEngine;
    function ParseModList (p: PChar; var ModList: Pointer): boolean;

var
    (* Compiler symbols *)
    SetCompilerOptions: TSetCompilerOptionsProc;
    CompileFile: TCompileFileProc;
    VersionDecode: TVersionDecodeProc;
    (* BuilderKit symbols *)
    GetCompilerErrors: TGetCompilerErrorsProc;
    GetCompilerErrorInfo: TGetCompilerErrorInfoProc;
    GetCompilerStatus: TGetCompilerStatusProc;
    SetDllName: TSetDllNameProc;
    ModListInit: TModListInitProc;
    AddModule: TAddModuleProc;
    SetModuleOption: TSetModuleOptionProc;
    GetLinkerError: TGetLinkerErrorProc;
    GetLinkerStatus: TGetLinkerStatusProc;
    (* Linker symbols *)
    LinkDll: TLinkDllProc;
    LinkDynDll: TLinkDllProc;
    LinkExe: TLinkExeProc;

Implementation

uses
    SysUtils,
    HostDynLoader;

var
    Module: TDynLoadModule;
    res: integer;

procedure InitSymbols;
begin
    (* Compiler symbols *)
    @CompileFile := Nil;
    @SetCompilerOptions := Nil;
    @VersionDecode := Nil;
    (* BuilderKit symbols *)
    @GetCompilerErrors := Nil;
    @GetCompilerErrorInfo := Nil;
    @GetCompilerStatus := Nil;
    @ModListInit := Nil;
    @AddModule := Nil;
    @SetModuleOption := Nil;
    @SetDllName := Nil;
    @GetLinkerError := Nil;
    @GetLinkerStatus := Nil;
    (* Linker symbols *)
    @LinkDll := Nil;
    @LinkDynDll := Nil;
    @LinkExe := Nil
end;

function LoadLib (libname: PChar; var M: TDynLoadModule): integer;
var
    res: integer;
    SymbolFound: boolean;
begin
    res := resFailed; // Failed to load library, default.
    M := TDynLoadModule.Create();
    if M.LoadModule (libname) then
    begin
        InitSymbols;
        SymbolFound := TRUE;
        (* Compiler symbols *)
        @SetCompilerOptions := M.GetModuleSymbolEx('SetCompilerOptions', SymbolFound);
        @CompileFile := M.GetModuleSymbolEx('CompileFile', SymbolFound);
        @VersionDecode := M.GetModuleSymbolEx('VersionDecode', SymbolFound);
        (* BuilderKit symbols *)
        @GetCompilerErrors := M.GetModuleSymbolEx('GetCompilerErrors', SymbolFound);
        @GetCompilerErrorInfo := M.GetModuleSymbolEx('GetCompilerErrorInfo', SymbolFound);
        @GetCompilerStatus := M.GetModuleSymbolEx('GetCompilerStatus', SymbolFound);
        @ModListInit := M.GetModuleSymbolEx('ModListInit', SymbolFound);
        @AddModule := M.GetModuleSymbolEx('AddModule', SymbolFound);
        @SetModuleOption := M.GetModuleSymbolEx('SetOption', SymbolFound);
        @SetDllName := M.GetModuleSymbolEx('SetDllName', SymbolFound);
        @GetLinkerError := M.GetModuleSymbolEx('GetLinkerError', SymbolFound);
        @GetLinkerStatus := M.GetModuleSymbolEx('GetLinkerStatus', SymbolFound);
        (* Linker symbols *)
        @LinkDll := M.GetModuleSymbolEx('LinkDll', SymbolFound);
        @LinkDynDll := M.GetModuleSymbolEx('LinkDynDll', SymbolFound);
    {$ifdef win32}
        @LinkExe := M.GetModuleSymbolEx('LinkExe', SymbolFound);
    {$endif}

        if SymbolFound then
            res := resSuccess // All symbols found, success.
        else
            res := resSymNotFound // Failed to find symbol(s). Incorrect version of library?
    end;
    Result := res
end;

function LoadEngine: integer;
begin
    Result := LoadLib (libCompPC, Module)
end;

procedure UnloadEngine;
begin
    Module.UnloadModule;
    Module.Free
end;

function ParseModList (p: PChar; var ModList: Pointer): boolean;
var
    res: boolean;
    i, k: integer;
    m: array [0..255] of char;
    opts: integer;
begin
    res := TRUE;
    ModListInit (ModList);
    i := 0; k := 0; opts := 0; m[0] := #0;
    while i < Length(p) DO
    begin
	if p[i] = '#' then
	    SetModuleOption (opts, isintf)
	else if p[i] = '*' then
	    SetModuleOption (opts, isdll)
	else if p[i] = '$' then
	    SetModuleOption (opts, ismain)
	else if p[i] = '+' then
	    SetModuleOption (opts, iskernel)
	else if p[i] = ' ' (* space is a delimiter *) then
    	    begin
		m[k] := #0; k := 0;
		//writeln('DEBUG: ', m);
		if m[0] <> #0 then
		    AddModule(m, opts, ModList)
		else
		    res := FALSE;
		opts := 0;
    	    end
	    else (* Letters *)
    	    begin
		m[k] := p[i]; Inc(k)
	    end;
	    m[k] := #0; Inc(i)
	end;
	if k <> 0 then
	begin
	    //writeln('DEBUG: ', m);
	    if m[0] <> #0 then
		AddModule(m, opts, ModList)
	    else
		res := FALSE
	end;
	Result := res
end;

end.
