(*
    Project: Component Pascal command-line compiler
    Organization: Oberon Revival team
    License: BSD style license
    References: http://oberonrevivalsourceforge.net/
*)

{$IFDEF FPC}
    {$mode objfpc}
{$ENDIF}

uses
    SysUtils, Strings,
    cdev, prj, cmsg;

Const
    CPver = 'Component Pascal command-line compiler';

var 
    isEngineLoaded,
    isCPErrListLoaded: boolean;
    Project: TProject;
    Mode: TMode;
    strBuf: PChar;

procedure TerminateWithMsg (msg: string; errno: integer);
begin
    WriteLn (msg);
    Halt (errno)
end;

procedure Help;
begin
    TerminateWithMsg ('Usage: cpcc <project_file.prj>', 100)
end;

procedure CompilerLogSuccess (compdata: TCompData);
begin
    WriteLn (Format('Compiled. pc=%d, dsize=%d', [compdata.pc, compdata.dsize]))
end;

procedure CompilerLogErrors (errors: Integer; comperr: TCompError; IOerr: integer);
    var ermes: string; uknown: boolean;
begin
    if IOerr = 0 then // if no I/O errors, it compiler's errors
    begin
        WriteLn (Format('%s(%d,%d) Error #%d: %s',
            [ExtractFileName (Project.Module), comperr.row, comperr.col, comperr.No, ErrorMessage (comperr.No)]));
        WriteLn (Format('%d error(s) detected', [errors]))
    end
    else
    begin
        Write ('Error: ');
        uknown := FALSE;
        case IOerr of
            RES_INVALID_PARAMETER: ermes := 'Invalid parameter (name or locator)';
            RES_NOT_FOUND: ermes := 'Location or file not found';
            RES_IO_ERROR: ermes := 'I/O error';
            RES_ACCESS_DENIED: ermes := 'Access denied';
            else
            begin
                uknown := TRUE;
                WriteLn ('#', IOerr)
            end
        end;
        if Not uknown then
            WriteLn (ermes)
    end
end;

procedure CompilerLogBegin (name: string);
begin
    WriteLn (Format('Compiling %s ...', [name]))
end;

procedure LinkerLogBegin;
begin
    WriteLn('Linking ...')
end;

procedure ReportBuildFailed (err: ErrType);
begin
    WriteLn('Linking failed.');
    WriteLn(Format('Error: %s', [err.mes]))
end;

procedure ReportBuildSuccess (codesize, filelen: integer);
begin
    WriteLn(Format('File %s written: %d %d', [Project.OutputFile, filelen, codesize]))
end;

procedure Initialize;
var
    res: integer;
    Msg: string;
begin
    res := LoadEngine;
    if res <> resSuccess then
    begin
        case res of
        resFailed:
            Msg := Format ('Failed to load library %s.', [libCompPC]);
        resSymNotFound:
            Msg := Format ('Invalid version of library %s', [libCompPC]);
        else
            Msg := Format ('Library %s not loaded. Error #%d', [libCompPC, res])
        end;
        TerminateWithMsg (Msg, 102)
    end
    else
        isEngineLoaded := TRUE;

    isCPErrListLoaded := LoadCPErrList;
    if Not isCPErrListLoaded then
        TerminateWithMsg (cmsgError, 201);
    strBuf := Strings.StrAlloc (512);
end;

procedure Finalize;
begin
    StrDispose (strBuf);
    if isEngineLoaded then
    begin
        UnloadEngine;
        if isCPErrListLoaded then
            UnloadCPErrList
    end
end;

procedure GetPar;
    var pathToProject: string;
begin
    if ParamCount = 0 then Help;
    pathToProject := ParamStr (1);
    if  (ReadProject (pathToProject, Project) <> 0)
        or (Project.Module = '')
        or (Project.LinkedMods = '')
        or (Project.OutputFile = '') then
        TerminateWithMsg ('Cannot read project definition from file. See documentation.', 101)
end;

procedure StringToPChar (St: String; var P: PChar);
begin
    if P <> Nil then
    begin
        if Strings.StrPCopy (P, St) <> P then
            P := Nil
    end
end;

procedure SetProjectPar;
begin
    if Project.LinkerMode = 'Dll' then
        Mode := ModeDll
    else if Project.LinkerMode = 'DynDll' then
        Mode := ModeDynDll
    else if Project.LinkerMode = 'Exe' then
        Mode := ModeExe
    else
        Mode := ModeNone;
    StringToPChar (Project.OutputFile, strBuf);
    SetDllName (strBuf);
end;

procedure CheckProjectPar (param: string; var modulesList: pointer);
var res: boolean;
begin
    if Mode = ModeNone then
        TerminateWithMsg ('Error: Invalid linker mode. Note: parameter is case-sensitive. See documentation.', 103);
    
    StringToPChar (param, strBuf);
    res := ParseModList (strBuf, modulesList);
    if Not res Or (modulesList = Nil) then
        TerminateWithMsg ('Error: Parameters list is empty or incorrect. See documentation.', 104)
end;

function Link (modList, resList: pointer; commandLine: boolean): boolean;
var res: boolean;
begin
    res := FALSE;
    case Mode of
        ModeDll: res := LinkDll (modList, resList);
        ModeDynDll: res := LinkDynDll (modList, resList);
        ModeExe: res := LinkExe (modList, resList, commandLine);
    end;
    Result := res
end;

procedure CompileAndLink (ModuleFile: string);
var
    ioerr: integer;
    modulesList: pointer;
    comperr: TCompError;
    compdata: TCompData;
    err: ErrType;
    codesize, filelen: integer;
begin
    CompilerLogBegin (ModuleFile);
    ioerr := -1;
    StringToPChar (ModuleFile, strBuf);
    if CompileFile (strBuf, ioerr) then
    begin
        GetCompilerStatus (compdata);
        CompilerLogSuccess (compdata);
        
        CheckProjectPar (Project.LinkedMods, modulesList);
        LinkerLogBegin;
        if Not Link (modulesList, Nil {No resource list}, FALSE {commandLine disabled}) then
        begin
            GetLinkerError (err);
            ReportBuildFailed (err)
        end
        else
        begin
            GetLinkerStatus (codesize, filelen);
            ReportBuildSuccess (codesize, filelen)
        end
    end
    else
    begin
        GetCompilerErrorInfo (comperr);
        CompilerLogErrors (GetCompilerErrors (), comperr, ioerr);
    end
end;

begin
    WriteLn (CPver);
    GetPar;
    Initialize;
    SetProjectPar;
    CompileAndLink (Project.Module);
    Finalize
end.
